; ==========================
; SECRET OF MANA CRAB SPRITE
;	by Kickchon
; ==========================

; ==================
; SPRITE TABLE USAGE
; ==================

; $C2,x				the crabs hitpoints (1)
; $1504,x			the pincers sprite number
; $1510,x
; $151C,x	
; $1528,x
; $1540,x	autodec
; $154C,x	autodec
; $1558,x	autodec

; $1534,x			used by kick_engine
; $1570,x			used by kick_engine
; $157C,x			used by kick_engine
; $1594,x			used by kick_engine
; $15AC,x	autodec		used by kick_engine
; $1602,x			used by kick_engine
; $160E,x			used by kick_engine
; $163E,x	autodec		used by kick_engine
; $187B,x			used by kick_engine
; $1FD6,x			used by kick_engine

; ========================
; KICKCHON's SPRITE ENGINE
; ========================

incsrc sprites/kick_engine_top.asm

; ================
; SPRITE CONSTANTS
; ================

!CRAB_MOVETIME	=	$0F
!CRAB_HITPOINTS	=	$01

!CRAB_HITPOINT	=	$C2
!CRAB_PINCERS	=	$1504

; ==================================
; KICKCHON's SPRITE ENGINE CONSTANTS
; ==================================

; The states the sprite may be in, should be an enumeration starting from $00

!STATE_ACCEL	=	$00
!STATE_MOVE	=	$01
!STATE_THINK	=	$02
!STATE_THROW	=	$03
!STATE_WAIT	=	$04
!STATE_CATCH	=	$05
!STATE_HIT	=	$06
!STATE_DEAD	=	$07

; The table for the state durations, only relevant if automatic state change is activated for the state

STATE_TIMES:	db	$10, $80, $30, $14, $1D, $14, $30, $FF

; The table which defines if the automatic state change should be used, $00 means no automatic change, $01 equals automatic change after time runs out

STATE_USETIMES:	db	$01, $01, $01, $01, $01, $01, $01, $00

; The table for the sprite's state processing functions, (main, init, finalize and death state)

STATE_FUNCS:	dw	SPRITE_ACCEL, SPRITE_MOVE, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_DEAD
STATE_INITFNCS:	dw	SPRITE_ACCELI, SPRITE_MOVEI, SPRITE_THINKI, SPRITE_THROWI, SPRITE_WAITI, SPRITE_CATCHI, SPRITE_HITI, SPRITE_DEADI
STATE_FINALFNC:	dw	SPRITE_ACCELF, SPRITE_MOVEF, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY, SPRITE_EMPTY
!STATE_DEADFNC	=	!STATE_DEAD

; The Animations the sprite has, should be an enumeration starting from $00

!ANIM_MOVE0	=	$00
!ANIM_MOVE1	=	$01
!ANIM_HIT	=	$02
!ANIM_THROW	=	$03

; Tile render mode, set to $02 for 16x16 tiles, $00 for 8x8 tiles

!GFX_TILEFMT	=	$02

; Table for the amount of tiles in the tilemaps, needs to be amount of tiles * 3

GFX_TILECNT:	db	$09, $09, $06, $0C

; The tilemaps for the sprite, format: <tile>, <xoffset>, <yoffset>, [<tile>, <xoffset>, <yoffset>, ...][, optional flipped values (those are not counted)]

GFX_MOVE0:	db	$C0,$FD,$00, $C2,$0D,$00, $AA,$FE,$F0, $AA,$02,$F0, $C2,$F3,$00, $C0,$03,$00
GFX_MOVE1:	db	$C4,$FB,$00, $C6,$0B,$00, $8E,$FD,$F0, $8E,$03,$F0, $C6,$F5,$00, $C4,$05,$00
GFX_HIT:	db	$C8,$F5,$00, $CE,$05,$00, $CE,$FB,$00, $C8,$0B,$00
GFX_THROW:	db	$CA,$F7,$FC, $CC,$07,$FC, $EA,$F7,$0C, $EC,$07,$0C, $CC,$F9,$FC, $CA,$09,$FC, $EC,$F9,$0C, $EA,$09,$0C

; The animation table

GFX_ANIMATIONS:	dw	GFX_MOVE0, GFX_MOVE1, GFX_HIT, GFX_THROW

; =============================
; GLOBAL SPRITE INITIALIZIATION
; =============================

GLOBAL_INIT:	LDA #!CRAB_HITPOINTS
		STA !CRAB_HITPOINT,x
		STZ !SPRITE_XACCEL,x
		LDA !SPRITE_NUMCUS,x
		INC A
		STA !CRAB_PINCERS,x
		%initialState(#!STATE_THINK)
		%moveSpriteGravity()
		RTS

; =====================
; COMMON STATE FUNCTION
; =====================

COMMON_MAIN:	LDA !SPRITE_STATEC,x
		CMP #!STATE_HIT
		BNE MAIN_CONTINUE
		RTS
MAIN_CONTINUE:	%handleSpriteContact()
		%jumpNoSpriteContact(MAIN_MARIO)
		%flipSpeedX()
		%flipAccelX()
		%flipDirectionBit()
		%updateAnimationDirection()
MAIN_MARIO:	%jumpNoMarioContact(RETURN)
		%jumpMarioHittingAbove(MAIN_LOSE)
		%jumpMarioHasStar(MAIN_LOSE)
		%jumpMarioSliding(MAIN_LOSE)
		%hurtMario()
		RTS
MAIN_LOSE:	%setNextState(#!STATE_HIT)
		%changeToNextState()
RETURN:		RTS

; =====================================
; SPRITE STATE INITIALIZATION FUNCTIONS
; =====================================

SPRITE_ACCELI:	LDX !SPRITE_CINDEX
		%setNextState(#!STATE_MOVE)
		%updateDirectionBit()
		LDA !SPRITE_FRAMET
		AND #!CRAB_MOVETIME
		BEQ ACCELI_SWAP
		%updateAnimationDirection()
		BRA ACCELI_ACCEL
ACCELI_SWAP:	LDA !SPRITE_ANIMID,x
		EOR #$01
		%changeAnimationA()
ACCELI_ACCEL:	%jumpDirectionBitSet(ACCELI_RIGHT)
		%decAccelX()
		RTS
ACCELI_RIGHT:	%incAccelX()
		RTS

SPRITE_MOVEI:	LDX !SPRITE_CINDEX
		%setNextState(#!STATE_THINK)
		RTS

SPRITE_THINKI:	LDX !SPRITE_CINDEX
		%changeAnimation(#!ANIM_MOVE0)
		LDA !MARIO_XPOSLO
		AND #$03
		BEQ THINKI_THROW
		%setNextState(#!STATE_ACCEL)
		RTS
THINKI_THROW:	%setNextState(#!STATE_THROW)
		RTS

SPRITE_THROWI:	LDX !SPRITE_CINDEX
		%updateDirectionBit()
		%setNextState(#!STATE_WAIT)
		%changeAnimation(#!ANIM_THROW)
		%jumpDirectionBitSet(THROWI_RIGHT)
		LDA !CRAB_PINCERS,x
		%spawnCustomSpriteRelativeA(#$F8,#$00)
		RTS
THROWI_RIGHT:	LDA !CRAB_PINCERS,x
		%spawnCustomSpriteRelativeA(#$08,#$00)
		RTS

SPRITE_WAITI:	LDX !SPRITE_CINDEX
		%setNextState(#!STATE_CATCH)
		%changeAnimation(#!ANIM_MOVE0)
		RTS

SPRITE_CATCHI:	LDX !SPRITE_CINDEX
		%setNextState(#!STATE_THINK)
		%changeAnimation(#!ANIM_THROW)
		RTS

SPRITE_HITI:	LDX !SPRITE_CINDEX
		%jumpMarioHasStar(HIT_STARDEATH)
		%jumpMarioSliding(HIT_DEATH)
		%handleMarioContact()
		%jumpNoSpinJumpOrYoshi(HIT_NODEATH)
		%showContactStar()
		%playSoundEffect(!SOUND_CHAN0,#$08)
		%changeSMWSpriteStatus(#!STATUS_SPNJMP)
		%useDefaultRoutine()
		RTS
HIT_STARDEATH:	LDA #$FF
		STA !CRAB_HITPOINT,x
HIT_DEATH:	%changeToDeathState()
		RTS
HIT_NODEATH:	LDA !CRAB_HITPOINT,x
		BEQ HIT_DEATH
		DEC !CRAB_HITPOINT,x
		%givePoints()
		%setNextState(#!STATE_THINK)
		%changeAnimation(#!ANIM_HIT)
		RTS

SPRITE_DEADI:	LDX !SPRITE_CINDEX
		%stayDead()
		%setAccelY(#$04)
		LDA !CRAB_HITPOINT,x
		CMP #$FF
		BEQ DEADI_STAR
		%givePoints()
		%jumpExternalDeath(DEADI_EXT)
		%clearSpeedX()
		%setSpeedY(#$20)
		RTS
DEADI_STAR:	%giveStarPoints()
DEADI_EXT:	%setSpeedY(#$D0)
		%setSpeedX(#$10)
		%jumpMarioRightSide(DEADI_FLIP)
		RTS
DEADI_FLIP:	%flipSpeedX()
		RTS

; ===================================
; SPRITE STATE FINALIZATION FUNCTIONS
; ===================================

SPRITE_ACCELF:	LDX !SPRITE_CINDEX
		%setAccelX(#$00)
		RTS

SPRITE_MOVEF:	LDX !SPRITE_CINDEX
		%setSpeedX(#$00)
		RTS

; ======================
; SPRITE STATE FUNCTIONS
; ======================

SPRITE_ACCEL:	LDX !SPRITE_CINDEX	
		LDA !SPRITE_FRAMET
		AND #!CRAB_MOVETIME
		BEQ ACCEL_SWAP
		%updateAnimationDirection()
		BRA ACCEL_AIR
ACCEL_SWAP:	LDA !SPRITE_ANIMID,x
		EOR #$01
		%changeAnimationA()
ACCEL_AIR:	%jumpExtraBitClear(ACCEL_WALL)
		%jumpNotInAir(ACCEL_WALL)
		BRA ACCEL_FLIP
ACCEL_WALL:	%jumpNotTouchingWall(ACCEL_MOVE)
ACCEL_FLIP:	%flipSpeedX()
		%flipAccelX()
		%flipDirectionBit()
		%updateAnimationDirection()
ACCEL_MOVE:	%updateSpeedX()
		%moveSpriteGravity()
		RTS

SPRITE_MOVE:	LDX !SPRITE_CINDEX
		LDA !SPRITE_FRAMET
		AND #!CRAB_MOVETIME
		BEQ MOVE_SWAP
		%updateAnimationDirection()
		BRA MOVE_AIR
MOVE_SWAP:	LDA !SPRITE_ANIMID,x
		EOR #$01
		%changeAnimationA()
MOVE_AIR:	%jumpExtraBitClear(MOVE_WALL)
		%jumpNotInAir(MOVE_WALL)
		BRA MOVE_FLIP
MOVE_WALL:	%jumpNotTouchingWall(MOVE_MOVE)
MOVE_FLIP:	%flipSpeedX()
		%flipDirectionBit()
		%updateAnimationDirection()
MOVE_MOVE:	%moveSpriteGravity()
		RTS

SPRITE_DEAD:	LDX !SPRITE_CINDEX
		LDA !SPRITE_YSPEED,x
		EOR #$80
		STA !CRAB_HITPOINT,x
		%updateSpeedY()
		AND !CRAB_HITPOINT,x
		AND #$80
		BEQ DEAD_MOVE
		%flipAccelY()
		%updateSpeedY()
		%clearAccelY()
DEAD_MOVE:	%moveSpriteNoGravity()
		RTS

; ====================
; EMPTY STATE FUNCTION
; ====================

SPRITE_EMPTY:	LDX !SPRITE_CINDEX
		RTS

; ========================
; KICKCHON's SPRITE ENGINE
; ========================

incsrc sprites/kick_engine_bottom.asm